import click

from gravity import process_manager


@click.command("status")
@click.pass_context
def cli(ctx):
    """Shut down process manager."""
    with process_manager.process_manager(state_dir=ctx.parent.state_dir, start_daemon=False) as pm:
        pm.shutdown()



import click

from gravity import config_manager
from gravity import options


@click.command("reregister")
@options.required_config_arg(name="old_config")
@options.required_config_arg(name="new_config", exists=True)
@click.pass_context
def cli(ctx, old_config, new_config):
    """Update path of registered config file.

    aliases: rename
    """
    with config_manager.config_manager(state_dir=ctx.parent.state_dir) as cm:
        cm.rename(old_config, new_config)



""" Galaxy Process Management superclass and utilities
"""
import contextlib
import errno
import hashlib
import logging
import os
import xml.etree.ElementTree as elementtree
from os import pardir
from os.path import abspath, dirname, exists, expanduser, isabs, join
from typing import Union

from yaml import safe_load

from gravity import __version__
from gravity.settings import Settings
from gravity.io import debug, error, exception, info, warn
from gravity.state import (
    ConfigFile,
    GravityState,
    service_for_service_type,
)
from gravity.util import recursive_update, yaml_safe_load_with_include

log = logging.getLogger(__name__)

DEFAULT_JOB_CONFIG_FILE = "config/job_conf.xml"
DEFAULT_STATE_DIR = join("~", ".config", "galaxy-gravity")
if "XDG_CONFIG_HOME" in os.environ:
    DEFAULT_STATE_DIR = join(os.environ["XDG_CONFIG_HOME"], "galaxy-gravity")


@contextlib.contextmanager
def config_manager(state_dir=None, python_exe=None):
    yield ConfigManager(state_dir=state_dir, python_exe=python_exe)


class ConfigManager(object):
    galaxy_server_config_section = "galaxy"
    gravity_config_section = "gravity"

    def __init__(self, state_dir=None, python_exe=None):
        if state_dir is None:
            state_dir = DEFAULT_STATE_DIR
        self.state_dir = abspath(expanduser(state_dir))
        debug(f"Gravity state dir: {self.state_dir}")
        self.config_state_path = join(self.state_dir, "configstate.yaml")
        self.python_exe = python_exe
        try:
            os.makedirs(self.state_dir)
        except OSError as exc:
            if exc.errno != errno.EEXIST:
                raise
        self.__convert_config()

    def __copy_config(self, old_path):
        with GravityState.open(old_path) as state:
            state.set_name(self.config_state_path)
        # copies on __exit__

    def __convert_config(self):
        config_state_json = join(self.state_dir, "configstate.json")
        if exists(config_state_json) and not exists(self.config_state_path):
            warn(f"Converting {config_state_json} to {self.config_state_path}")
            json_state = GravityState.open(config_state_json)
            self.__copy_config(config_state_json)
            assert exists(self.config_state_path), f"Conversion failed ({self.config_state_path} does not exist)"
            yaml_state = GravityState.open(self.config_state_path)
            assert json_state == yaml_state, f"Converted config differs from previous config, remove {self.config_state_path} to retry"
            os.unlink(config_state_json)

    def get_config(self, conf, defaults=None):
        defaults = defaults or {}
        server_section = self.galaxy_server_config_section
        with open(conf) as config_fh:
            config_dict = yaml_safe_load_with_include(config_fh)
        _gravity_config = config_dict.get(self.gravity_config_section) or {}
        gravity_config = Settings(**recursive_update(defaults, _gravity_config))
        if gravity_config.log_dir is None:
            gravity_config.log_dir = join(expanduser(self.state_dir), "log")

        if server_section not in config_dict and self.gravity_config_section not in config_dict:
            error(f"Config file {conf} does not look like valid Galaxy, Reports or Tool Shed configuration file")
            return None

        app_config = config_dict.get(server_section) or {}

        config = ConfigFile()
        config.attribs = {}
        config.services = []
        config.instance_name = gravity_config.instance_name
        config.config_type = server_section
        config.attribs["galaxy_infrastructure_url"] = app_config.get("galaxy_infrastructure_url", "").rstrip("/")
        if gravity_config.tusd.enable and not config.attribs["galaxy_infrastructure_url"]:
            exception("To run the tusd server you need to set galaxy_infrastructure_url in the galaxy section of galaxy.yml")
        config.attribs["app_server"] = gravity_config.app_server
        config.attribs["log_dir"] = gravity_config.log_dir
        config.attribs["virtualenv"] = gravity_config.virtualenv
        config.attribs["gunicorn"] = gravity_config.gunicorn.dict()
        config.attribs["tusd"] = gravity_config.tusd.dict()
        config.attribs["celery"] = gravity_config.celery.dict()
        config.attribs["handlers"] = gravity_config.handlers
        # Store gravity version, in case we need to convert old setting
        config.attribs['gravity_version'] = __version__
        webapp_service_names = []

        # shortcut for galaxy configs in the standard locations -- explicit arg ?
        config.attribs["galaxy_root"] = app_config.get("root") or gravity_config.galaxy_root
        if config.attribs["galaxy_root"] is None:
            if os.environ.get("GALAXY_ROOT_DIR"):
                config.attribs["galaxy_root"] = abspath(os.environ["GALAXY_ROOT_DIR"])
            elif exists(join(dirname(conf), pardir, "lib", "galaxy")):
                config.attribs["galaxy_root"] = abspath(join(dirname(conf), pardir))
            elif conf.endswith(join('galaxy', 'config', 'sample', 'galaxy.yml.sample')):
                config.attribs["galaxy_root"] = abspath(join(dirname(conf), pardir, pardir, pardir, pardir))
            else:
                exception(f"Cannot locate Galaxy root directory: set $GALAXY_ROOT_DIR or `root' in the `galaxy' section of {conf}")

        if gravity_config.gunicorn.enable:
            if config.attribs["gunicorn"]["preload"] is None:
                config.attribs["gunicorn"]["preload"] = config.attribs["app_server"] != "unicornherder"
            config.services.append(service_for_service_type(config.attribs["app_server"])(config_type=config.config_type))
        if gravity_config.celery.enable:
            config.services.append(service_for_service_type("celery")(config_type=config.config_type))
        if gravity_config.celery.enable_beat:
            config.services.append(service_for_service_type("celery-beat")(config_type=config.config_type))
        if gravity_config.tusd.enable:
            config.services.append(service_for_service_type("tusd")(config_type=config.config_type))

        if not app_config.get("job_config_file") and app_config.get("job_config"):
            # config embedded directly in Galaxy config
            job_config = app_config["job_config"]
        else:
            # If this is a Galaxy config, parse job_conf.xml for any *static* standalone handlers
            # TODO: use galaxy config parsing ?
            job_config = app_config.get("job_config_file", DEFAULT_JOB_CONFIG_FILE)
            if not isabs(job_config):
                # FIXME: relative to root
                job_config = abspath(join(config.attribs["galaxy_root"], job_config))
                if not exists(job_config):
                    job_config = None
        if config.config_type == "galaxy" and job_config:
            for service_name in [x["service_name"] for x in ConfigManager.get_job_config(job_config) if x["service_name"] not in webapp_service_names]:
                config.services.append(service_for_service_type("standalone")(config_type=config.config_type, service_name=service_name))

        # Dynamic job handlers are configured using `job_handler_count` in galaxy.yml.
        #
        # FIXME: This should imply explicit configuration of the handler assignment method. If not explicitly set, the
        # web process will be a handler, which is not desirable when dynamic handlers are used. Currently Gravity
        # doesn't parse that part of the job config. See logic in lib/galaxy/web_stack/handlers.py _get_is_handler() to
        # see how this is determined.
        self.create_handler_services(gravity_config, config)
        self.create_gxit_services(gravity_config, app_config, config)
        return config

    def create_handler_services(self, gravity_config: Settings, config):
        expanded_handlers = self.expand_handlers(gravity_config, config)
        for service_name, handler_settings in expanded_handlers.items():
            pools = handler_settings.get('pools')
            config.services.append(
                service_for_service_type("standalone")(config_type=config.config_type, service_name=service_name, server_pools=pools))

    def create_gxit_services(self, gravity_config: Settings, app_config, config):
        if app_config.get("interactivetools_enable") and gravity_config.gx_it_proxy.enable:
            # TODO: resolve against data_dir, or bring in galaxy-config ?
            # CWD in supervisor template is galaxy_root, so this should work for simple cases as is
            gxit_config = gravity_config.gx_it_proxy
            gxit_config.sessions = app_config.get("interactivetools_map", gxit_config.sessions)
            gxit_config.verbose = '--verbose' if gxit_config.verbose else ''
            config.services.append(service_for_service_type("gx-it-proxy")(config_type=config.config_type))
        config.attribs["gx_it_proxy"] = gravity_config.gx_it_proxy.dict()

    @staticmethod
    def expand_handlers(gravity_config: Settings, config):
        handlers = gravity_config.handlers or {}
        expanded_handlers = {}
        default_name_template = "{name}_{process}"
        for service_name, handler_config in handlers.items():
            count = handler_config.get("processes", 1)
            name_template = handler_config.get("name_template")
            if name_template is None:
                if count == 1 and service_name[-1].isdigit():
                    # Assume we have an explicit handler name, don't apply pattern
                    expanded_handlers[service_name] = handler_config
                    continue
            name_template = (name_template or default_name_template).strip()
            for index in range(count):
                expanded_service_name = name_template.format(name=service_name, process=index, instance_name=config.instance_name)
                if expanded_service_name not in expanded_handlers:
                    expanded_handlers[expanded_service_name] = handler_config
        return expanded_handlers

    @staticmethod
    def get_job_config(conf: Union[str, dict]):
        """Extract handler names from job_conf.xml"""
        # TODO: use galaxy job conf parsing
        rval = []
        if isinstance(conf, str):
            if conf.endswith('.xml'):
                root = elementtree.parse(conf).getroot()
                for handler in (root.find("handlers") or []):
                    rval.append({"service_name": handler.attrib["id"]})
                return rval
            elif conf.endswith(('.yml', '.yaml')):
                with open(conf) as job_conf_fh:
                    conf = safe_load(job_conf_fh.read())
            else:
                exception(f"Unknown job config file type: {conf}")
        if isinstance(conf, dict):
            handling = conf.get('handling') or {}
            processes = handling.get('processes') or {}
            for handler_name, handler_options in processes.items():
                rval.append({
                    "service_name": handler_name,
                    "environment": (handler_options or {}).get("environment", None)
                })
        return rval

    def _register_config_file(self, key, val):
        """Persist a newly added config file, or update (overwrite) the value
        of a previously persisted config.
        """
        with self.state as state:
            state.config_files[key] = val

    def _deregister_config_file(self, key):
        """Deregister a previously registered config file.  The caller should
        ensure that it was previously registered.
        """
        with self.state as state:
            if "remove_configs" not in state:
                state.remove_configs = {}
            state.remove_configs[key] = state.config_files.pop(key)

    def _purge_config_file(self, key):
        """Forget a previously deregister config file.  The caller should
        ensure that it was previously deregistered.
        """
        with self.state as state:
            del state.remove_configs[key]
            if not state.remove_configs:
                del state["remove_configs"]

    def determine_config_changes(self):
        """The magic: Determine what has changed since the last time.

        Caller should pass the returned config to register_config_changes to persist.
        """
        # 'update' here is synonymous with 'add or update'
        instances = set()
        new_configs = {}
        meta_changes = {"changed_instances": set(), "remove_instances": [], "remove_configs": self.get_remove_configs()}
        for config_file, stored_config in self.get_registered_configs().items():
            new_config = stored_config
            try:
                ini_config = self.get_config(config_file, defaults=stored_config.defaults)
            except OSError as exc:
                warn("Unable to read %s (hint: use `rename` or `remove` to fix): %s", config_file, exc)
                new_configs[config_file] = stored_config
                instances.add(stored_config["instance_name"])
                continue
            if ini_config["instance_name"] is not None:
                # instance name is explicitly set in the config
                instance_name = ini_config["instance_name"]
                if ini_config["instance_name"] != stored_config["instance_name"]:
                    # instance name has changed
                    # (removal of old instance will happen later if no other config references it)
                    new_config["update_instance_name"] = instance_name
                meta_changes["changed_instances"].add(instance_name)
            else:
                # instance name is dynamically generated
                instance_name = stored_config["instance_name"]
            if ini_config["attribs"] != stored_config["attribs"]:
                new_config["update_attribs"] = ini_config["attribs"]
                meta_changes["changed_instances"].add(instance_name)
            # make sure this instance isn't removed
            instances.add(instance_name)
            services = []
            for service in ini_config["services"]:
                for stored_service in stored_config["services"]:
                    if service.full_match(stored_service):
                        # service is configured and has no changes
                        break
                else:
                    # instance has a new service or service has config change
                    if "update_services" not in new_config:
                        new_config["update_services"] = []
                    new_config["update_services"].append(service)
                    meta_changes["changed_instances"].add(instance_name)
                # make sure this service isn't removed
                services.append(service)
            for service in stored_config["services"]:
                if service not in services:
                    if "remove_services" not in new_config:
                        new_config["remove_services"] = []
                    new_config["remove_services"].append(service)
                    meta_changes["changed_instances"].add(instance_name)
            new_configs[config_file] = new_config
        # once finished processing all configs, find any instances which have been deleted
        for instance_name in self.get_registered_instances(include_removed=True):
            if instance_name not in instances:
                meta_changes["remove_instances"].append(instance_name)
        return new_configs, meta_changes

    def register_config_changes(self, configs, meta_changes):
        """Persist config changes to the JSON state file. When a config
        changes, a process manager may perform certain actions based on these
        changes. This method can be called once the actions are complete.
        """
        for config_file in meta_changes["remove_configs"].keys():
            self._purge_config_file(config_file)
        for config_file, config in configs.items():
            if "update_attribs" in config:
                config["attribs"] = config.pop("update_attribs")
            if "update_instance_name" in config:
                config["instance_name"] = config.pop("update_instance_name")
            if "update_services" in config or "remove_services" in config:
                remove = config.pop("remove_services", [])
                services = config.pop("update_services", [])
                # need to prevent old service defs from overwriting new ones
                for service in config["services"]:
                    if service not in remove and service not in services:
                        services.append(service)
                config["services"] = services
            self._register_config_file(config_file, config)

    @property
    def state(self):
        """Public property to access persisted config state"""
        return GravityState.open(self.config_state_path)

    @property
    def instance_count(self):
        """The number of configured instances"""
        return len(self.state.config_files)

    @property
    def single_instance(self):
        """Indicate if there is only one configured instance"""
        return self.instance_count == 1

    def get_registered_configs(self, instances=None):
        """Return the persisted values of all config files registered with the config manager."""
        configs = self.state.config_files
        if instances is not None:
            for config_file, config in list(configs.items()):
                if config["instance_name"] not in instances:
                    configs.pop(config_file)
        return configs

    def get_remove_configs(self):
        """Return the persisted values of all config files pending removal by the process manager."""
        return self.state.get("remove_configs", {})

    def get_registered_config(self, config_file):
        """Return the persisted value of the named config file."""
        return self.state.config_files.get(config_file, None)

    def get_registered_instances(self, include_removed=False):
        """Return the persisted names of all instances across all registered configs."""
        rval = []
        configs = list(self.state.config_files.values())
        if include_removed:
            configs.extend(list(self.get_remove_configs().values()))
        for config in configs:
            if config["instance_name"] not in rval:
                rval.append(config["instance_name"])
        return rval

    def get_instance_config(self, instance_name):
        for config in list(self.state.config_files.values()):
            if config["instance_name"] == instance_name:
                return config
        exception(f'Instance "{instance_name}" unknown, known instance(s) are {", ".join(self.get_registered_instance_names())}.')

    def get_registered_instance_names(self):
        return [c['instance_name'] for c in self.state.config_files.values()]

    def get_instance_services(self, instance_name):
        return self.get_instance_config(instance_name)["services"]

    def get_registered_services(self):
        rval = []
        for config_file, config in self.state.config_files.items():
            for service in config["services"]:
                service["config_file"] = config_file
                service["instance_name"] = config["instance_name"]
                rval.append(service)
        return rval

    def auto_register(self):
        """Attempt to automatically register a config file if none are registered."""
        if self.instance_count == 0:
            if os.environ.get("GALAXY_CONFIG_FILE"):
                configs = [os.environ["GALAXY_CONFIG_FILE"]]
            else:
                configs = (os.path.join("config", "galaxy.yml"), os.path.join("config", "galaxy.yml.sample"))
            for config in configs:
                if exists(config):
                    # This should always be the case if instance_count == 0
                    if not self.is_registered(abspath(config)):
                        self.add([config])
                    return

    def is_registered(self, config_file):
        return config_file in self.get_registered_configs()

    def add(self, config_files, galaxy_root=None):
        """Public method to add (register) config file(s)."""
        for config_file in config_files:
            config_file = abspath(expanduser(config_file))
            if self.is_registered(config_file):
                warn("%s is already registered", config_file)
                continue
            defaults = None
            if galaxy_root is not None:
                defaults = {"galaxy_root": galaxy_root}
            conf = self.get_config(config_file, defaults=defaults)
            if conf is None:
                exception(f"Cannot add {config_file}: File is unknown type")
            if conf["instance_name"] is None:
                conf["instance_name"] = conf["config_type"] + "-" + hashlib.md5(os.urandom(32)).hexdigest()[:12]
            conf_data = {
                "config_type": conf["config_type"],
                "instance_name": conf["instance_name"],
                "attribs": conf["attribs"],
                "services": [],  # services will be populated by the update method
            }
            self._register_config_file(config_file, conf_data)
            info("Registered %s config: %s", conf["config_type"], config_file)

    def rename(self, old, new):
        if not self.is_registered(old):
            error("%s is not registered", old)
            return
        conf = self.get_config(new)
        if conf is None:
            exception(f"Cannot add {new}: File is unknown type")
        with self.state as state:
            state.config_files[new] = state.config_files.pop(old)
        info("Reregistered config %s as %s", old, new)

    def remove(self, config_files):
        # FIXME: paths are checked by click now
        # allow the arg to be instance names
        configs_by_instance = self.get_registered_configs(instances=config_files)
        if configs_by_instance:
            supplied_config_files = []
            config_files = list(configs_by_instance.keys())
        else:
            supplied_config_files = [abspath(cf) for cf in config_files]
            config_files = []
        for config_file in supplied_config_files:
            if not self.is_registered(config_file):
                warn("%s is not registered", config_file)
            else:
                config_files.append(config_file)
        for config_file in config_files:
            self._deregister_config_file(config_file)
            info("Deregistered config: %s", config_file)



import click

from gravity import config_manager
from gravity import options


@click.command("register")
@options.required_config_arg(exists=True, nargs=-1)
@click.pass_context
def cli(ctx, config):
    """Register config file(s).

    aliases: add
    """
    with config_manager.config_manager(state_dir=ctx.parent.state_dir) as cm:
        cm.add(config)



import click

from gravity import config_manager


@click.command("configs")
@click.pass_context
def cli(ctx):
    """List registered config files.

    aliases: list
    """
    with config_manager.config_manager(state_dir=ctx.parent.state_dir) as cm:
        registered = cm.get_registered_configs()
        if registered:
            click.echo("%-12s  %-24s  %s" % ("TYPE", "INSTANCE NAME", "CONFIG PATH"))
            for config in sorted(registered.keys()):
                click.echo("%-12s  %-24s  %s" % (registered[config].get("config_type", "unknown"), registered[config].get("instance_name", "unknown"), config))
        else:
            click.echo("No config files registered")



import click

from gravity import options
from gravity import process_manager


@click.command("follow")
@options.required_instance_arg()
@click.pass_context
def cli(ctx, instance):
    """Follow log files of configured instances.

    If INSTANCE matches an instance name, logs of all services configured for the instance are followed.

    If INSTANCE does not match an instance name, it is assumed to be a service and only logs of the listed service(s)
    are followed."""
    with process_manager.process_manager(state_dir=ctx.parent.state_dir, start_daemon=False) as pm:
        pm.follow(instance)



import click

from gravity import process_manager


@click.command("update")
@click.option("--force", is_flag=True, help="Force rewriting of process config files")
@click.pass_context
def cli(ctx, force):
    """Update process manager from config changes."""
    with process_manager.process_manager(state_dir=ctx.parent.state_dir, start_daemon=False) as pm:
        pm.update(force)



"""
"""
import errno
import os
import shlex
import shutil
import subprocess
import time
from os.path import exists, join

from gravity.io import debug, error, exception, info, warn
from gravity.process_manager import BaseProcessManager
from gravity.state import GracefulMethod
from gravity.util import which

from supervisor import supervisorctl  # type: ignore

DEFAULT_SUPERVISOR_SOCKET_PATH = os.environ.get("SUPERVISORD_SOCKET", '%(here)s/supervisor.sock')

SUPERVISORD_CONF_TEMPLATE = f""";
; This file is maintained by Galaxy - CHANGES WILL BE OVERWRITTEN
;

[unix_http_server]
file = {DEFAULT_SUPERVISOR_SOCKET_PATH}

[supervisord]
logfile = %(here)s/supervisord.log
pidfile = %(here)s/supervisord.pid
loglevel = info
nodaemon = false

[rpcinterface:supervisor]
supervisor.rpcinterface_factory = supervisor.rpcinterface:make_main_rpcinterface

[supervisorctl]
serverurl = unix://{DEFAULT_SUPERVISOR_SOCKET_PATH}

[include]
files = supervisord.conf.d/*.d/*.conf supervisord.conf.d/*.conf
"""

# TODO: with more templating you only need one of these
SUPERVISORD_SERVICE_TEMPLATES = {}
SUPERVISORD_SERVICE_TEMPLATES["unicornherder"] = """;
; This file is maintained by Galaxy - CHANGES WILL BE OVERWRITTEN
;

[program:{program_name}]
command         = {command}
directory       = {galaxy_root}
umask           = {galaxy_umask}
autostart       = true
autorestart     = true
startsecs       = 15
stopwaitsecs    = 65
environment     = {environment}
numprocs        = 1
stdout_logfile  = {log_file}
redirect_stderr = true
{process_name_opt}
"""

SUPERVISORD_SERVICE_TEMPLATES["gunicorn"] = """;
; This file is maintained by Galaxy - CHANGES WILL BE OVERWRITTEN
;

[program:{program_name}]
command         = {command}
directory       = {galaxy_root}
umask           = {galaxy_umask}
autostart       = true
autorestart     = true
startsecs       = 15
stopwaitsecs    = 65
environment     = {environment}
numprocs        = 1
stdout_logfile  = {log_file}
redirect_stderr = true
{process_name_opt}
"""

SUPERVISORD_SERVICE_TEMPLATES["celery"] = """;
; This file is maintained by Galaxy - CHANGES WILL BE OVERWRITTEN
;

[program:{program_name}]
command         = {command}
directory       = {galaxy_root}
umask           = {galaxy_umask}
autostart       = true
autorestart     = true
startsecs       = 10
stopwaitsecs    = 10
environment     = {environment}
numprocs        = 1
stdout_logfile  = {log_file}
redirect_stderr = true
{process_name_opt}
"""

SUPERVISORD_SERVICE_TEMPLATES["celery-beat"] = """;
; This file is maintained by Galaxy - CHANGES WILL BE OVERWRITTEN
;

[program:{program_name}]
command         = {command}
directory       = {galaxy_root}
umask           = {galaxy_umask}
autostart       = true
autorestart     = true
startsecs       = 10
stopwaitsecs    = 10
environment     = {environment}
numprocs        = 1
stdout_logfile  = {log_file}
redirect_stderr = true
{process_name_opt}
"""

SUPERVISORD_SERVICE_TEMPLATES["tusd"] = """;
; This file is maintained by Galaxy - CHANGES WILL BE OVERWRITTEN
;
[program:{program_name}]
command         = {command}
directory       = {galaxy_root}
umask           = {galaxy_umask}
autostart       = true
autorestart     = true
startsecs       = 10
stopwaitsecs    = 10
environment     = {environment}
numprocs        = 1
stdout_logfile  = {log_file}
redirect_stderr = true
{process_name_opt}
"""

SUPERVISORD_SERVICE_TEMPLATES["gx-it-proxy"] = """;
; This file is maintained by Galaxy - CHANGES WILL BE OVERWRITTEN
;

[program:{program_name}]
command         = {command}
directory       = {galaxy_root}
umask           = {galaxy_umask}
autostart       = true
autorestart     = true
startsecs       = 10
stopwaitsecs    = 10
environment     = {environment}
numprocs        = 1
stdout_logfile  = {log_file}
redirect_stderr = true
{process_name_opt}
"""

SUPERVISORD_SERVICE_TEMPLATES["standalone"] = """;
; This file is maintained by Galaxy - CHANGES WILL BE OVERWRITTEN
;

[program:{program_name}]
command         = {command}
directory       = {galaxy_root}
autostart       = true
autorestart     = true
startsecs       = 20
stopwaitsecs    = 65
environment     = {environment}
numprocs        = 1
stdout_logfile  = {log_file}
redirect_stderr = true
{process_name_opt}
"""

SUPERVISORD_GROUP_TEMPLATE = """;
; This file is maintained by Galaxy - CHANGES WILL BE OVERWRITTEN
;

[group:{instance_name}]
programs = {programs}
"""


class SupervisorProcessManager(BaseProcessManager):
    def __init__(self, state_dir=None, start_daemon=True, foreground=False):
        super(SupervisorProcessManager, self).__init__(state_dir=state_dir)
        self.supervisord_exe = which("supervisord")
        self.supervisor_state_dir = join(self.state_dir, "supervisor")
        self.supervisord_conf_path = join(self.supervisor_state_dir, "supervisord.conf")
        self.supervisord_conf_dir = join(self.supervisor_state_dir, "supervisord.conf.d")
        self.supervisord_pid_path = join(self.supervisor_state_dir, "supervisord.pid")
        self.supervisord_sock_path = os.environ.get("SUPERVISORD_SOCKET", join(self.supervisor_state_dir, "supervisor.sock"))
        self.__supervisord_popen = None
        self.foreground = foreground

        if not exists(self.supervisord_conf_dir):
            os.makedirs(self.supervisord_conf_dir)

        if start_daemon:
            self.__supervisord()

    @property
    def use_group(self):
        return not self.config_manager.single_instance

    @property
    def log_file(self):
        return join(self.supervisor_state_dir, "supervisord.log")

    def __supervisord_is_running(self):
        try:
            assert exists(self.supervisord_pid_path)
            assert exists(self.supervisord_sock_path)
            os.kill(int(open(self.supervisord_pid_path).read()), 0)
            return True
        except Exception:
            return False

    def __supervisord(self):
        format_vars = {"supervisor_state_dir": self.supervisor_state_dir, "supervisord_conf_dir": self.supervisord_conf_dir}
        supervisord_cmd = [self.supervisord_exe, "-c", self.supervisord_conf_path]
        if self.foreground:
            supervisord_cmd.append('--nodaemon')
        if not self.__supervisord_is_running():
            # any time that supervisord is not running, let's rewrite supervisord.conf
            open(self.supervisord_conf_path, "w").write(SUPERVISORD_CONF_TEMPLATE.format(**format_vars))
            self.__supervisord_popen = subprocess.Popen(supervisord_cmd, env=os.environ)
            rc = self.__supervisord_popen.poll()
            if rc:
                error("supervisord exited with code %d" % rc)
            # FIXME: don't wait forever
            while not exists(self.supervisord_pid_path) or not exists(self.supervisord_sock_path):
                debug(f"Waiting for {self.supervisord_pid_path}")
                time.sleep(0.5)

    def __get_supervisor(self):
        """Return the supervisor proxy object

        Should probably use this more rather than supervisorctl directly
        """
        options = supervisorctl.ClientOptions()
        options.realize(args=["-c", self.supervisord_conf_path])
        return supervisorctl.Controller(options).get_supervisor()

    def terminate(self):
        if self.foreground:
            # if running in foreground, if terminate is called, then supervisord should've already received a SIGINT
            self.__supervisord_popen and self.__supervisord_popen.wait()

    def _service_program_name(self, instance_name, service):
        if self.use_group:
            return f"{instance_name}_{service['config_type']}_{service['service_type']}_{service['service_name']}"
        else:
            return service["service_name"]

    def __update_service(self, config_file, config, attribs, service, instance_conf_dir, instance_name):
        if self.use_group:
            process_name_opt = f"process_name    = {service['service_name']}"
        else:
            process_name_opt = ""

        program_name = self._service_program_name(instance_name, service)

        # used by the "standalone" service type
        attach_to_pool_opt = ""
        server_pools = service.get("server_pools")
        if server_pools:
            _attach_to_pool_opt = " ".join(f"--attach-to-pool={server_pool}" for server_pool in server_pools)
            # Insert a single leading space
            attach_to_pool_opt = f" {_attach_to_pool_opt}"

        virtualenv_dir = attribs.get("virtualenv")
        virtualenv_bin = f'{os.path.join(virtualenv_dir, "bin")}{os.path.sep}' if virtualenv_dir else ""
        gunicorn_options = attribs["gunicorn"].copy()
        gunicorn_options["preload"] = "--preload" if gunicorn_options["preload"] else ""

        format_vars = {
            "log_dir": attribs["log_dir"],
            "log_file": self._service_log_file(attribs["log_dir"], program_name),
            "config_type": service["config_type"],
            "server_name": service["service_name"],
            "attach_to_pool_opt": attach_to_pool_opt,
            "gunicorn": gunicorn_options,
            "celery": attribs["celery"],
            "galaxy_infrastructure_url": attribs["galaxy_infrastructure_url"],
            "tusd": attribs["tusd"],
            "gx_it_proxy": attribs["gx_it_proxy"],
            "galaxy_umask": service.get("umask", "022"),
            "program_name": program_name,
            "process_name_opt": process_name_opt,
            "galaxy_conf": config_file,
            "galaxy_root": attribs["galaxy_root"],
            "virtualenv_bin": virtualenv_bin,
            "supervisor_state_dir": self.supervisor_state_dir,
            "state_dir": self.state_dir,
        }
        format_vars["command"] = service.command_template.format(**format_vars)
        conf = join(instance_conf_dir, f"{service['config_type']}_{service['service_type']}_{service['service_name']}.conf")

        if not exists(attribs["log_dir"]):
            os.makedirs(attribs["log_dir"])

        template = SUPERVISORD_SERVICE_TEMPLATES.get(service["service_type"])
        if not template:
            raise Exception(f"Unknown service type: {service['service_type']}")

        environment = self._service_environment(service, attribs)
        format_vars["environment"] = ",".join("{}={}".format(k, shlex.quote(v.format(**format_vars))) for k, v in environment.items())

        with open(conf, "w") as out:
            out.write(template.format(**format_vars))

    def _process_config_changes(self, configs, meta_changes, force=False):
        # remove the services of any configs which have been removed
        for config in meta_changes["remove_configs"].values():
            instance_name = config["instance_name"]
            instance_conf_dir = join(self.supervisord_conf_dir, f"{instance_name}.d")
            for service in config["services"]:
                info("Removing service %s", self._service_program_name(instance_name, service))
                conf = join(instance_conf_dir, f"{service['config_type']}_{service['service_type']}_{service['service_name']}.conf")
                if exists(conf):
                    os.unlink(conf)

        # update things for existing or new configs
        for config_file, config in configs.items():
            instance_name = config["instance_name"]
            attribs = config["attribs"]
            update_all_configs = False or force

            # config attribs have changed (galaxy_root, virtualenv, etc.)
            if "update_attribs" in config:
                info(f"Updating all dependent services of config {config_file} due to changes")
                attribs = config["update_attribs"]
                update_all_configs = True

            # instance name has changed, so supervisor group config must change
            if "update_instance_name" in config:
                instance_name = config["update_instance_name"]
                info("Creating new instance for name change: %s -> %s", config["instance_name"], instance_name)
                update_all_configs = True

            # always attempt to make the config dir
            instance_conf_dir = join(self.supervisord_conf_dir, f"{instance_name}.d")
            try:
                os.makedirs(instance_conf_dir)
            except OSError as exc:
                if exc.errno != errno.EEXIST:
                    raise

            if update_all_configs:
                for service in config["services"]:
                    info("Updating service %s", self._service_program_name(instance_name, service))
                    self.__update_service(config_file, config, attribs, service, instance_conf_dir, instance_name)

            # new services
            if "update_services" in config:
                for service in config["update_services"]:
                    info("Creating or updating service %s", self._service_program_name(instance_name, service))
                    self.__update_service(config_file, config, attribs, service, instance_conf_dir, instance_name)

            # deleted services
            if "remove_services" in config:
                for service in config["remove_services"]:
                    info("Removing service %s", self._service_program_name(instance_name, service))
                    conf = join(instance_conf_dir, f"{service['config_type']}_{service['service_type']}_{service['service_name']}.conf")
                    if exists(conf):
                        os.unlink(conf)

            # sanity check, make sure everything that should exist does exist
            for service in config["services"]:
                conf = join(instance_conf_dir, f"{service['config_type']}_{service['service_type']}_{service['service_name']}.conf")
                if service not in config.get("remove_services", []) and not exists(conf):
                    self.__update_service(config_file, config, attribs, service, instance_conf_dir, instance_name)
                    warn(f"Missing service config recreated: {conf}")

        # all configs referencing an instance name have been removed (or their
        # instance names have changed), nuke the group
        for instance_name in meta_changes["remove_instances"]:
            info("Removing instance %s", instance_name)
            instance_conf_dir = join(self.supervisord_conf_dir, f"{instance_name}.d")
            if exists(instance_conf_dir):
                shutil.rmtree(instance_conf_dir)
            conf = join(self.supervisord_conf_dir, f"group_{instance_name}.conf")
            if exists(conf):
                os.unlink(join(conf))

        # persist to the state file
        self.config_manager.register_config_changes(configs, meta_changes)

        # now we can create/update the instance group
        for instance_name in meta_changes["changed_instances"]:
            programs = []
            for service in self.config_manager.get_registered_services():
                if service["instance_name"] == instance_name and service["service_type"] != "uwsgi":
                    programs.append(f"{instance_name}_{service['config_type']}_{service['service_type']}_{service['service_name']}")
            conf = join(self.supervisord_conf_dir, f"group_{instance_name}.conf")
            if programs and self.use_group:
                format_vars = {"instance_conf_dir": instance_conf_dir, "instance_name": instance_name, "programs": ",".join(programs)}
                open(conf, "w").write(SUPERVISORD_GROUP_TEMPLATE.format(**format_vars))
            else:
                # no programs for the group, so it should be removed
                if exists(conf):
                    os.unlink(conf)

    def __start_stop(self, op, instance_names):
        self.update()
        instance_names, service_names, registered_instance_names = self.get_instance_names(instance_names)
        for instance_name in instance_names:
            target = f"{instance_name}:*" if self.use_group else "all"
            self.supervisorctl(op, target)
            for service in self.config_manager.get_instance_services(instance_name):
                if service["service_type"] == "uwsgi":
                    self.supervisorctl(op, f"{instance_name}_{service['config_type']}_{service['service_name']}")
        # shortcut for just passing service names directly
        for name in service_names:
            self.supervisorctl(op, name)

    def __reload_graceful(self, op, instance_names):
        self.update()
        instance_names, service_names, registered_instance_names = self.get_instance_names(instance_names)
        if not instance_names:
            instance_names = registered_instance_names
        known_services = []
        unknown_services = list(service_names)
        for instance_name in instance_names:
            for service in self.config_manager.get_instance_services(instance_name):
                program_name = self._service_program_name(instance_name, service)
                known_services.append(program_name)
                if service_names:
                    if program_name not in service_names:
                        continue
                    else:
                        unknown_services.remove(program_name)
                if service.graceful_method == GracefulMethod.SIGHUP:
                    self.supervisorctl("signal", "SIGHUP", program_name)
                else:
                    self.supervisorctl("restart", program_name)
        if unknown_services:
            exception(f'Invalid service(s): {", ".join(unknown_services)}. Known service(s) are {", ".join(known_services)}')

    def start(self, instance_names):
        self.__start_stop("start", instance_names)
        self.supervisorctl("status")

    def stop(self, instance_names):
        self.__start_stop("stop", instance_names)
        # Exit supervisor if all processes are stopped
        supervisor = self.__get_supervisor()
        proc_infos = supervisor.getAllProcessInfo()
        if all([i["state"] == 0 for i in proc_infos]):
            info("All processes stopped, supervisord will exit")
            self.shutdown()
        else:
            info("Not all processes stopped, supervisord not shut down (hint: see `galaxyctl status`)")

    def restart(self, instance_names):
        self.__start_stop("restart", instance_names)

    def reload(self, instance_names):
        self.__reload_graceful("reload", instance_names)

    def graceful(self, instance_names):
        self.__reload_graceful("graceful", instance_names)

    def status(self):
        # TODO: create our own formatted output
        # supervisor = self.get_supervisor()
        # all_infos = supervisor.getAllProcessInfo()
        self.supervisorctl("status")

    def shutdown(self):
        self.supervisorctl("shutdown")
        while self.__supervisord_is_running():
            debug("Waiting for supervisord to terminate")
            time.sleep(0.5)
        info("supervisord has terminated")

    def update(self, force=False):
        """Add newly defined servers, remove any that are no longer present"""
        configs, meta_changes = self.config_manager.determine_config_changes()
        self._process_config_changes(configs, meta_changes, force)
        # only need to update if supervisord is running, otherwise changes will be picked up at next start
        if self.__supervisord_is_running():
            self.supervisorctl("update")

    def supervisorctl(self, *args, **kwargs):
        if not self.__supervisord_is_running():
            warn("supervisord is not running")
            return
        try:
            supervisorctl.main(args=["-c", self.supervisord_conf_path] + list(args))
        except SystemExit as e:
            # supervisorctl.main calls sys.exit(), so we catch that
            if e.code == 0:
                pass
            else:
                raise



""" Click definitions for various shared options and arguments.
"""
import click


def debug_option():
    return click.option("-d", "--debug", is_flag=True, help="Enables debug mode.")


def state_dir_option():
    return click.option(
        "--state-dir", type=click.Path(file_okay=False, writable=True, resolve_path=True), help="Where process management configs and state will be stored."
    )


def no_log_option():
    return click.option(
        '--quiet', is_flag=True, default=False, help="Only output supervisor logs, do not include process logs"
    )


def required_config_arg(name="config", exists=False, nargs=None):
    arg_type = click.Path(
        exists=exists,
        file_okay=True,
        dir_okay=False,
        readable=True,
        resolve_path=True,
    )
    if nargs is None:
        return click.argument(name, type=arg_type)
    else:
        return click.argument(name, nargs=nargs, type=arg_type)


def required_instance_arg():
    return click.argument("instance", nargs=-1)



import click

from gravity import config_manager, options
from gravity import process_manager
from gravity.io import info, exception


@click.command("start")
@options.required_instance_arg()
@click.option("-f", "--foreground", is_flag=True, default=False, help="Run in foreground")
@options.no_log_option()
@click.pass_context
def cli(ctx, foreground, instance, quiet=False):
    """Start configured services.

    If INSTANCE matches an instance name, all services configured for the instance are started.

    If INSTANCE does not match an instance name, it is assumed to be a service and only the listed service(s) are
    started."""
    if not instance:
        with config_manager.config_manager(state_dir=ctx.parent.state_dir) as cm:
            # If there are no configs registered, we will attempt to auto-register one
            cm.auto_register()
        if not cm.instance_count:
            exception(
                "Nothing to start: no Galaxy instances configured and no Galaxy configuration files found, "
                "see `galaxyctl register --help`")
    with process_manager.process_manager(state_dir=ctx.parent.state_dir, foreground=foreground) as pm:
        pm.start(instance)
        if foreground:
            pm.follow(instance, quiet=quiet)
        elif pm.config_manager.single_instance:
            config = list(pm.config_manager.get_registered_configs().values())[0]
            info(f"Log files are in {config.attribs['log_dir']}")
        else:
            for config in pm.config_manager.get_registered_configs(instances=instance or None).values():
                info(f"Log files for {config.instance_name} are in {config.attribs['log_dir']}")



""" Galaxy Process Management superclass and utilities
"""

import contextlib
import importlib
import inspect
import os
import subprocess
from abc import ABCMeta, abstractmethod

from gravity.config_manager import ConfigManager
from gravity.io import exception
from gravity.util import which


# If at some point we have additional process managers we can make a factory,
# but for the moment there's only supervisor.
@contextlib.contextmanager
def process_manager(*args, **kwargs):
    # roulette!
    for filename in os.listdir(os.path.dirname(__file__)):
        if filename.endswith(".py") and not filename.startswith("_"):
            mod = importlib.import_module("gravity.process_manager." + filename[: -len(".py")])
            for name in dir(mod):
                obj = getattr(mod, name)
                if not name.startswith("_") and inspect.isclass(obj) and issubclass(obj, BaseProcessManager) and obj != BaseProcessManager:
                    pm = obj(*args, **kwargs)
                    try:
                        yield pm
                    finally:
                        pm.terminate()
                    return


class BaseProcessManager(object, metaclass=ABCMeta):

    def __init__(self, state_dir=None, start_daemon=True, foreground=False):
        self.config_manager = ConfigManager(state_dir=state_dir)
        self.state_dir = self.config_manager.state_dir
        self.tail = which("tail")

    def _service_log_file(self, log_dir, program_name):
        return os.path.join(log_dir, program_name + ".log")

    def _service_program_name(self, instance_name, service):
        return f"{instance_name}_{service['config_type']}_{service['service_type']}_{service['service_name']}"

    def _service_environment(self, service, attribs):
        environment = service.get_environment()
        environment_from = service.environment_from
        if not environment_from:
            environment_from = service.service_type
        environment.update(attribs.get(environment_from, {}).get("environment", {}))
        return environment

    @abstractmethod
    def start(self, instance_names):
        """ """

    @abstractmethod
    def _process_config_changes(self, configs, meta_changes):
        """ """

    @abstractmethod
    def terminate(self):
        """ """

    @abstractmethod
    def stop(self, instance_names):
        """ """

    @abstractmethod
    def restart(self, instance_names):
        """ """

    @abstractmethod
    def reload(self, instance_names):
        """ """

    def follow(self, instance_names, quiet=False):
        # supervisor has a built-in tail command but it only works on a single log file. `galaxyctl supervisorctl tail
        # ...` can be used if desired, though
        if not self.tail:
            exception("`tail` not found on $PATH, please install it")
        instance_names, service_names, registered_instance_names = self.get_instance_names(instance_names)
        log_files = []
        if quiet:
            cmd = [self.tail, "-f", self.log_file]
            tail_popen = subprocess.Popen(cmd)
            tail_popen.wait()
        else:
            if not instance_names:
                instance_names = registered_instance_names
            for instance_name in instance_names:
                config = self.config_manager.get_instance_config(instance_name)
                log_dir = config["attribs"]["log_dir"]
                if not service_names:
                    services = self.config_manager.get_instance_services(instance_name)
                    for service in services:
                        program_name = self._service_program_name(instance_name, service)
                        log_files.append(self._service_log_file(log_dir, program_name))
                else:
                    log_files.extend([self._service_log_file(log_dir, s) for s in service_names])
                cmd = [self.tail, "-f"] + log_files
                tail_popen = subprocess.Popen(cmd)
                tail_popen.wait()

    @abstractmethod
    def graceful(self, instance_names):
        """ """

    @abstractmethod
    def update(self, instance_names):
        """ """

    @abstractmethod
    def shutdown(self, instance_names):
        """ """

    def get_instance_names(self, instance_names):
        registered_instance_names = self.config_manager.get_registered_instances()
        unknown_instance_names = []
        if instance_names:
            _instance_names = []
            for n in instance_names:
                if n in registered_instance_names:
                    _instance_names.append(n)
                else:
                    unknown_instance_names.append(n)
            instance_names = _instance_names
        elif registered_instance_names:
            instance_names = registered_instance_names
        else:
            exception("No instances registered (hint: `galaxyctl register /path/to/galaxy.yml`)")
        return instance_names, unknown_instance_names, registered_instance_names



import sys

import click

from gravity import config_manager
from gravity import options
from gravity.io import exception


@click.command("show")
@options.required_config_arg(exists=True)
@click.pass_context
def cli(ctx, config):
    """Show details of registered config.

    aliases: get
    """
    with config_manager.config_manager(state_dir=ctx.parent.state_dir) as cm:
        config_data = cm.get_registered_config(config)
        if config_data is None:
            newline = "\n"
            error_message = f"{config} is not a registered config file.{newline}"
            registered_configs = cm.get_registered_configs()
            if registered_configs:
                registered_configs_str = "\n".join(registered_configs)
                error_message = f'{error_message}Registered config files are:{newline}{registered_configs_str}{newline}{newline}'
            else:
                error_message = f'{error_message}No config files have been registered.{newline}{newline}'
            error_message = f'{error_message}To register this config file run "galaxyctl register {config}".'
            exception(error_message)
        else:
            config_data.dump(sys.stdout)



import sys
import traceback

import click


DEBUG = False


def debug(message, *args):
    if args:
        message = message % args
    if DEBUG:
        click.echo(message)


def info(message, *args):
    if args:
        message = message % args
    click.echo(click.style(message, bold=True, fg="green"))


def error(message, *args):
    if args:
        message = message % args
    if DEBUG and sys.exc_info()[0] is not None:
        click.echo(traceback.format_exc(), nl=False)
    click.echo(click.style(message, bold=True, fg="red"), err=True)


def warn(message, *args):
    if args:
        message = message % args
    click.echo(click.style(message, fg="red"), err=True)


def exception(message):
    raise click.ClickException(click.style(message, bold=True, fg="red"))



""" Command line utilities for managing Galaxy servers
"""

import os

import click

from gravity import io
from gravity import options


# FIXME: -p/--python-exe unimplemented
CONTEXT_SETTINGS = {
    "auto_envvar_prefix": "GRAVITY",
    "help_option_names": ["-h", "--help"]
}

# FIXME: incomplete aliases
COMMAND_ALIASES = {
    "list": "configs",
    "add": "register",
    "remove": "deregister",
    "forget": "deregister",
    "get": "show",
    "reload": "graceful",
}


cmd_folder = os.path.abspath(os.path.join(os.path.dirname(__file__), "commands"))


def set_debug(debug_opt):
    if debug_opt:
        io.DEBUG = True


def list_cmds():
    rv = []
    for filename in os.listdir(cmd_folder):
        if filename.endswith(".py") and filename.startswith("cmd_"):
            rv.append(filename[len("cmd_"): -len(".py")])
    rv.sort()
    return rv


def name_to_command(name):
    try:
        mod_name = "gravity.commands.cmd_" + name
        mod = __import__(mod_name, None, None, ["cli"])
    except ImportError as e:
        io.error(f"Problem loading command {name}, exception {e}")
        return
    return mod.cli


class GravityCLI(click.MultiCommand):
    def list_commands(self, ctx):
        return list_cmds()

    def get_command(self, ctx, name):
        if name in COMMAND_ALIASES:
            name = COMMAND_ALIASES[name]
        return name_to_command(name)


# Shortcut for running Galaxy in the foreground
@click.command(context_settings=CONTEXT_SETTINGS)
@options.debug_option()
@options.state_dir_option()
@options.no_log_option()
@click.pass_context
def galaxy(ctx, debug, state_dir, quiet):
    """Run Galaxy server in the foreground"""
    set_debug(debug)
    ctx.state_dir = state_dir
    mod = __import__("gravity.commands.cmd_start", None, None, ["cli"])
    return ctx.invoke(mod.cli, foreground=True, quiet=quiet)


@click.command(cls=GravityCLI, context_settings=CONTEXT_SETTINGS)
@options.debug_option()
@options.state_dir_option()
@click.pass_context
def galaxyctl(ctx, debug, state_dir):
    """Manage Galaxy server configurations and processes."""
    set_debug(debug)
    ctx.state_dir = state_dir



from enum import Enum
from typing import (
    Any,
    Dict,
    Optional,
)
from pydantic import BaseModel, BaseSettings, Extra, Field, validator


def none_to_default(cls, v, field):
    if all(
        (
            # Cater for the occasion where field.default in (0, False)
            getattr(field, "default", None) is not None,
            v is None,
        )
    ):
        return field.default
    else:
        return v


class LogLevel(str, Enum):
    debug = "DEBUG"
    info = "INFO"
    warning = "WARNING"
    error = "ERROR"


class AppServer(str, Enum):
    gunicorn = "gunicorn"
    unicornherder = "unicornherder"


class Pool(str, Enum):
    prefork = "prefork"
    eventlet = "eventlet"
    gevent = "gevent"
    solo = "solo"
    processes = "processes"
    threads = "threads"


class TusdSettings(BaseModel):
    enable: bool = Field(False, description="""
Enable tusd server.
If enabled, you also need to set up your proxy as outlined in https://docs.galaxyproject.org/en/latest/admin/nginx.html#receiving-files-via-the-tus-protocol.
""")
    tusd_path: str = Field(default="tusd", description="""Path to tusd binary""")
    host: str = Field("localhost", description="Host to bind the tusd server to")
    port: int = Field(1080, description="Port to bind the tusd server to")
    upload_dir: str = Field(description="""
Directory to store uploads in.
Must match ``tus_upload_store`` setting in ``galaxy:`` section.
""")
    hooks_enabled_events: str = Field(default="pre-create", description="""
Comma-separated string of enabled tusd hooks.

Leave at the default value to require authorization at upload creation time.
This means Galaxy's web process does not need to be running after creating the initial
upload request.

Set to empty string to disable all authorization. This means data can be uploaded (but not processed)
without the Galaxy web process being available.

You can find a list of available hooks at https://github.com/tus/tusd/blob/master/docs/hooks.md#list-of-available-hooks.
""")
    extra_args: str = Field(default="", description="Extra arguments to pass to tusd command line.")
    environment: Dict[str, str] = Field(
        default={},
        description="""
Extra environment variables and their values to set when running the service. A dictionary where keys are the variable
names.
""")


class CelerySettings(BaseModel):
    enable: bool = Field(True, description="Enable Celery distributed task queue.")
    enable_beat: bool = Field(True, description="Enable Celery Beat periodic task runner.")
    concurrency: int = Field(2, ge=0, description="Number of Celery Workers to start.")
    loglevel: LogLevel = Field(LogLevel.debug, description="Log Level to use for Celery Worker.")
    queues: str = Field("celery,galaxy.internal,galaxy.external", description="Queues to join")
    pool: Pool = Field(Pool.threads, description="Pool implementation")
    extra_args: str = Field(default="", description="Extra arguments to pass to Celery command line.")
    environment: Dict[str, str] = Field(
        default={},
        description="""
Extra environment variables and their values to set when running the service. A dictionary where keys are the variable
names.
""")

    class Config:
        use_enum_values = True


class GunicornSettings(BaseModel):
    enable: bool = Field(True, description="Enable Galaxy gunicorn server.")
    bind: str = Field(
        default="localhost:8080",
        description="The socket to bind. A string of the form: ``HOST``, ``HOST:PORT``, ``unix:PATH``, ``fd://FD``. An IP is a valid HOST.",
    )
    workers: int = Field(
        default=1,
        ge=1,
        description="""
Controls the number of Galaxy application processes Gunicorn will spawn.
Increased web performance can be attained by increasing this value.
If Gunicorn is the only application on the server, a good starting value is the number of CPUs * 2 + 1.
4-12 workers should be able to handle hundreds if not thousands of requests per second.
""")
    timeout: int = Field(
        default=300,
        ge=0,
        description="""
Gunicorn workers silent for more than this many seconds are killed and restarted.
Value is a positive number or 0. Setting it to 0 has the effect of infinite timeouts by disabling timeouts for all workers entirely.
If you disable the ``preload`` option workers need to have finished booting within the timeout.
""")
    extra_args: str = Field(default="", description="Extra arguments to pass to Gunicorn command line.")
    preload: Optional[bool] = Field(
        default=None,
        description="""
Use Gunicorn's --preload option to fork workers after loading the Galaxy Application.
Consumes less memory when multiple processes are configured. Default is ``false`` if using unicornherder, else ``true``.
""")
    environment: Dict[str, str] = Field(
        default={},
        description="""
Extra environment variables and their values to set when running the service. A dictionary where keys are the variable
names.
""")


class GxItProxySettings(BaseModel):
    enable: bool = Field(default=False, description="Set to true to start gx-it-proxy")
    ip: str = Field(default="localhost", description="Public-facing IP of the proxy")
    port: int = Field(default=4002, description="Public-facing port of the proxy")
    sessions: str = Field(
        default="database/interactivetools_map.sqlite",
        description="""
Routes file to monitor.
Should be set to the same path as ``interactivetools_map`` in the ``galaxy:`` section.
""")
    verbose: bool = Field(default=True, description="Include verbose messages in gx-it-proxy")
    forward_ip: Optional[str] = Field(
        default=None,
        description="""
Forward all requests to IP.
This is an advanced option that is only needed when proxying to remote interactive tool container that cannot be reached through the local network.
""")
    forward_port: Optional[int] = Field(
        default=None,
        description="""
Forward all requests to port.
This is an advanced option that is only needed when proxying to remote interactive tool container that cannot be reached through the local network.""")
    reverse_proxy: Optional[bool] = Field(
        default=False,
        description="""
Rewrite location blocks with proxy port.
This is an advanced option that is only needed when proxying to remote interactive tool container that cannot be reached through the local network.
""")
    environment: Dict[str, str] = Field(
        default={},
        description="""
Extra environment variables and their values to set when running the service. A dictionary where keys are the variable
names.
""")


class Settings(BaseSettings):
    """
    Configuration for Gravity process manager.
    ``uwsgi:`` section will be ignored if Galaxy is started via Gravity commands (e.g ``./run.sh``, ``galaxy`` or ``galaxyctl``).
    """

    galaxy_root: Optional[str] = Field(
        None,
        description="""
Specify Galaxy's root directory.
Gravity will attempt to find the root directory, but you can set the directory explicitly with this option.
""")
    log_dir: Optional[str] = Field(
        None,
        description="""
Set to a directory that should contain log files for the processes controlled by Gravity.
If not specified defaults to ``<state_dir>/log``.
""")
    virtualenv: Optional[str] = Field(None, description="""
Set to Galaxy's virtualenv directory.
If not specified, Gravity assumes all processes are on PATH.
""")
    app_server: AppServer = Field(
        AppServer.gunicorn,
        description="""
Select the application server.
``gunicorn`` is the default application server.
``unicornherder`` is a production-oriented manager for (G)unicorn servers that automates zero-downtime Galaxy server restarts,
similar to uWSGI Zerg Mode used in the past.
""")
    instance_name: str = Field(default="_default_", description="""Override the default instance name.
this is hidden from you when running a single instance.""")
    gunicorn: GunicornSettings = Field(default={}, description="Configuration for Gunicorn.")
    celery: CelerySettings = Field(default={}, description="Configuration for Celery Processes.")
    gx_it_proxy: GxItProxySettings = Field(default={}, description="Configuration for gx-it-proxy.")
    # The default value for tusd is a little awkward, but is a convenient way to ensure that if
    # a user enables tusd that they most also set upload_dir, and yet have the default be valid.
    tusd: TusdSettings = Field(default={'upload_dir': ''}, description="""
Configuration for tusd server (https://github.com/tus/tusd).
The ``tusd`` binary must be installed manually and made available on PATH (e.g in galaxy's .venv/bin directory).
""")
    handlers: Dict[str, Dict[str, Any]] = Field(
        default={},
        description="""
Configure dynamic handlers in this section.
See https://docs.galaxyproject.org/en/latest/admin/scaling.html#dynamically-defined-handlers for details.
""")

    # Use validators to turn None to default value
    _normalize_gunicorn = validator("gunicorn", allow_reuse=True, pre=True)(none_to_default)
    _normalize_gx_it_proxy = validator("gx_it_proxy", allow_reuse=True, pre=True)(none_to_default)
    _normalize_celery = validator("celery", allow_reuse=True, pre=True)(none_to_default)
    _normalize_tusd = validator("tusd", allow_reuse=True, pre=True)(none_to_default)

    class Config:
        env_prefix = "gravity_"
        env_nested_delimiter = "."
        case_sensitive = False
        use_enum_values = True
        # Ignore extra fields so you can switch from gravity versions that recognize new fields
        # to an older version that does not specify the fields, without having to comment them out.
        extra = Extra.ignore



import click

from gravity import options
from gravity import process_manager


@click.command("restart")
@options.required_instance_arg()
@click.pass_context
def cli(ctx, instance):
    """Restart configured services.

    If INSTANCE matches an instance name, all services configured for the instance are restarted.

    If INSTANCE does not match an instance name, it is assumed to be a service and only the listed service(s) are
    restarted."""
    with process_manager.process_manager(state_dir=ctx.parent.state_dir) as pm:
        pm.restart(instance)



import click

from gravity import config_manager
from gravity import options


@click.command("deregister")
@options.required_config_arg(nargs=-1)
@click.pass_context
def cli(ctx, config):
    """Deregister config file(s).

    aliases: remove, forget
    """
    with config_manager.config_manager(state_dir=ctx.parent.state_dir) as cm:
        cm.remove(config)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import ast
import os
import re

from setuptools import setup, find_packages


with open("README.rst") as file:
    long_description = file.read()

long_description += "\n\n"
with open("HISTORY.rst") as file:
    long_description += file.read()

with open(os.path.join("gravity", "__init__.py")) as f:
    init_contents = f.read()

    def get_var(var_name):
        pattern = re.compile(r"%s\s+=\s+(.*)" % var_name)
        match = pattern.search(init_contents).group(1)
        return str(ast.literal_eval(match))

    version = get_var("__version__")

setup(
    name="gravity",
    version=version,
    packages=find_packages(),
    description="Command-line utilities to assist in managing Galaxy servers",
    long_description=long_description,
    long_description_content_type="text/x-rst",
    url="https://github.com/galaxyproject/gravity",
    author="The Galaxy Team",
    author_email="team@galaxyproject.org",
    license="MIT",
    keywords="gravity galaxy",
    python_requires=">=3.6",
    install_requires=["Click", "supervisor", "pyyaml", "ruamel.yaml", "pydantic", "jsonref"],
    entry_points={"console_scripts": [
        "galaxy = gravity.cli:galaxy",
        "galaxyctl = gravity.cli:galaxyctl",
    ]},
    classifiers=[
        "Intended Audience :: System Administrators",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Operating System :: POSIX",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    zip_safe=False,
)



import click

from gravity import config_manager


@click.command("instances")
@click.pass_context
def cli(ctx):
    """List all known instances."""
    with config_manager.config_manager(state_dir=ctx.parent.state_dir) as cm:
        configs = cm.get_registered_configs()
        instances = cm.get_registered_instances()
        if instances:
            click.echo("%-24s  %-10s  %-10s  %s" % ("INSTANCE NAME", "TYPE", "SERVER", "NAME"))
            # not the most efficient...
            for instance in instances:
                instance_str = instance
                for config in configs.values():
                    if config["instance_name"] == instance:
                        for service in config["services"]:
                            click.echo("%-24s  %-10s  %-10s  %s" % (instance_str, service.config_type, service.service_type, service.service_name))
                            instance_str = ""
                if instance_str == instance:
                    click.echo("%-24s  no services configured" % instance)



""" Classes to represent and manipulate gravity's stored configuration and
state data.
"""
import enum
import errno
import os
import sys
from collections import defaultdict

import yaml

from gravity.util import AttributeDict


GALAXY_YML_SAMPLE_PATH = "lib/galaxy/config/sample/galaxy.yml.sample"
DEFAULT_GALAXY_ENVIRONMENT = {
    "PYTHONPATH": "lib",
    "GALAXY_CONFIG_FILE": "{galaxy_conf}",
}


class GracefulMethod(enum.Enum):
    DEFAULT = 0
    SIGHUP = 1


class Service(AttributeDict):
    service_type = "service"
    service_name = "_default_"
    environment_from = None
    default_environment = {}
    graceful_method = GracefulMethod.DEFAULT

    def __init__(self, *args, **kwargs):
        super(Service, self).__init__(*args, **kwargs)
        if "service_type" not in kwargs:
            self["service_type"] = self.__class__.service_type
        if "service_name" not in kwargs:
            self["service_name"] = self.__class__.service_name

    def __eq__(self, other):
        return self["config_type"] == other["config_type"] and self["service_type"] == other["service_type"] and self["service_name"] == other["service_name"]

    def full_match(self, other):
        return set(self.keys()) == set(other.keys()) and all([self[k] == other[k] for k in self if not k.startswith("_")])

    def get_environment(self):
        return self.default_environment.copy()


class GalaxyGunicornService(Service):
    service_type = "gunicorn"
    service_name = "gunicorn"
    default_environment = DEFAULT_GALAXY_ENVIRONMENT
    graceful_method = GracefulMethod.SIGHUP
    command_template = "{virtualenv_bin}gunicorn 'galaxy.webapps.galaxy.fast_factory:factory()'" \
                       " --timeout {gunicorn[timeout]}" \
                       " --pythonpath lib" \
                       " -k galaxy.webapps.galaxy.workers.Worker" \
                       " -b {gunicorn[bind]}" \
                       " --workers={gunicorn[workers]}" \
                       " --config python:galaxy.web_stack.gunicorn_config" \
                       " {gunicorn[preload]}" \
                       " {gunicorn[extra_args]}"

    def get_environment(self):
        # Works around https://github.com/galaxyproject/galaxy/issues/11821
        environment = self.default_environment.copy()
        if sys.platform == 'darwin':
            environment["OBJC_DISABLE_INITIALIZE_FORK_SAFETY"] = "YES"
        return environment


class GalaxyUnicornHerderService(Service):
    service_type = "unicornherder"
    service_name = "unicornherder"
    environment_from = "gunicorn"
    graceful_method = GracefulMethod.SIGHUP
    default_environment = DEFAULT_GALAXY_ENVIRONMENT
    command_template = "{virtualenv_bin}unicornherder --pidfile {supervisor_state_dir}/{program_name}.pid --" \
                       " 'galaxy.webapps.galaxy.fast_factory:factory()'" \
                       " --timeout {gunicorn[timeout]}" \
                       " --pythonpath lib" \
                       " -k galaxy.webapps.galaxy.workers.Worker" \
                       " -b {gunicorn[bind]}" \
                       " --workers={gunicorn[workers]}" \
                       " --config python:galaxy.web_stack.gunicorn_config" \
                       " {gunicorn[preload]}" \
                       " {gunicorn[extra_args]}"

    def get_environment(self):
        environment = self.default_environment.copy()
        if sys.platform == 'darwin':
            environment["OBJC_DISABLE_INITIALIZE_FORK_SAFETY"] = "YES"
        environment["GALAXY_CONFIG_LOG_DESTINATION"] = "{log_dir}/gunicorn.log"
        return environment


class GalaxyCeleryService(Service):
    service_type = "celery"
    service_name = "celery"
    default_environment = DEFAULT_GALAXY_ENVIRONMENT
    command_template = "{virtualenv_bin}celery" \
                       " --app galaxy.celery worker" \
                       " --concurrency {celery[concurrency]}" \
                       " --loglevel {celery[loglevel]}" \
                       " --pool {celery[pool]}" \
                       " --queues {celery[queues]}" \
                       " {celery[extra_args]}"


class GalaxyCeleryBeatService(Service):
    service_type = "celery-beat"
    service_name = "celery-beat"
    default_environment = DEFAULT_GALAXY_ENVIRONMENT
    command_template = "{virtualenv_bin}celery" \
                       " --app galaxy.celery" \
                       " beat" \
                       " --loglevel {celery[loglevel]}" \
                       " --schedule {state_dir}/celery-beat-schedule"


class GalaxyGxItProxyService(Service):
    service_type = "gx-it-proxy"
    service_name = "gx-it-proxy"
    default_environment = {
        "npm_config_yes": "true",
    }
    command_template = "{virtualenv_bin}npx gx-it-proxy --ip {gx_it_proxy[ip]} --port {gx_it_proxy[port]}" \
                       " --sessions {gx_it_proxy[sessions]} {gx_it_proxy[verbose]}"


class GalaxyTUSDService(Service):
    service_type = "tusd"
    service_name = "tusd"
    command_template = "{tusd[tusd_path]} -host={tusd[host]} -port={tusd[port]} -upload-dir={tusd[upload_dir]}" \
                       " -hooks-http={galaxy_infrastructure_url}/api/upload/hooks" \
                       " -hooks-http-forward-headers=X-Api-Key,Cookie {tusd[extra_args]}" \
                       " -hooks-enabled-events {tusd[hooks_enabled_events]}"


class GalaxyStandaloneService(Service):
    service_type = "standalone"
    service_name = "standalone"
    # FIXME: supervisor-specific
    command_template = "{virtualenv_bin}python ./lib/galaxy/main.py -c {galaxy_conf} --server-name={server_name}{attach_to_pool_opt}" \
                       " --pid-file={supervisor_state_dir}/{program_name}.pid"


class ConfigFile(AttributeDict):
    def __init__(self, *args, **kwargs):
        super(ConfigFile, self).__init__(*args, **kwargs)
        services = []
        for service in self.get("services", []):
            service_class = SERVICE_CLASS_MAP.get(service["service_type"], Service)
            services.append(service_class(**service))
        self.services = services

    @property
    def defaults(self):
        return {
            "instance_name": self["instance_name"],
            "galaxy_root": self["attribs"]["galaxy_root"],
            "log_dir": self["attribs"]["log_dir"],
            "gunicorn":  self.gunicorn_config,
        }

    @property
    def gunicorn_config(self):
        # We used to store bind_address and bind_port instead of a gunicorn config key, so restore from here
        gunicorn = self["attribs"].get("gunicorn")
        if not gunicorn and 'bind_address' in self["attribs"]:
            return {'bind': f'{self["attribs"]["bind_address"]}:{self["attribs"]["bind_port"]}'}
        return gunicorn


class GravityState(AttributeDict):
    @classmethod
    def open(cls, name):
        try:
            s = cls.loads(open(name).read())
        except (OSError, IOError) as exc:
            if exc.errno == errno.ENOENT:
                yaml.dump({}, open(name, "w"))
                s = cls()
        s._name = name
        return s

    def __init__(self, *args, **kwargs):
        super(GravityState, self).__init__(*args, **kwargs)
        normalized_state = defaultdict(dict)
        for key in ("config_files",):
            if key not in self:
                self[key] = {}
            for config_file, config_dict in self[key].items():
                # resolve path, so we always deal with absolute and symlink-resolved paths
                config_file = os.path.realpath(config_file)
                if config_file.endswith(GALAXY_YML_SAMPLE_PATH):
                    root_dir = config_dict['attribs']['galaxy_root']
                    non_sample_path = os.path.join(root_dir, 'config', 'galaxy.yml')
                    if os.path.exists(non_sample_path):
                        config_file = non_sample_path
                normalized_state[key][config_file] = ConfigFile(config_dict)
        self.update(normalized_state)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        with open(self._name, "w") as fh:
            self.dump(fh)

    def set_name(self, name):
        self._name = name


def service_for_service_type(service_type):
    try:
        return SERVICE_CLASS_MAP[service_type]
    except KeyError:
        raise RuntimeError(f"Unknown service type: {service_type}")


# TODO: better to pull this from __class__.service_type
SERVICE_CLASS_MAP = {
    "gunicorn": GalaxyGunicornService,
    "unicornherder": GalaxyUnicornHerderService,
    "celery": GalaxyCeleryService,
    "celery-beat": GalaxyCeleryBeatService,
    "gx-it-proxy": GalaxyGxItProxyService,
    "tusd": GalaxyTUSDService,
    "standalone": GalaxyStandaloneService,
}



"""
"""
import collections.abc
import copy
import os
import sys

import jsonref
import ruamel.yaml
import yaml
from gravity.settings import Settings


class SafeLoaderWithInclude(yaml.SafeLoader):
    def __init__(self, stream):
        self.__config_dir = os.path.dirname(stream.name)
        super().__init__(stream)

    def include(self, node):
        included = os.path.join(self.__config_dir, self.construct_scalar(node))
        with open(included) as fh:
            return yaml.load(fh, SafeLoaderWithInclude)


SafeLoaderWithInclude.add_constructor('!include', SafeLoaderWithInclude.include)


def yaml_safe_load_with_include(stream):
    return yaml.load(stream, SafeLoaderWithInclude)


class AttributeDict(dict):
    yaml_tag = "tag:yaml.org,2002:map"

    @classmethod
    def loads(cls, s, *args, **kwargs):
        return cls(yaml.safe_load(s, *args, **kwargs))

    @classmethod
    def to_yaml(cls, representer, node):
        d = {}
        for k in node.keys():
            if not k.startswith("_"):
                d[k] = node[k]
        return representer.represent_mapping(cls.yaml_tag, d)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._yaml = ruamel.yaml.YAML()
        self._yaml.register_class(self.__class__)

    def __eq__(self, other):
        return all([other[k] == v for k, v in self.items() if not k.startswith("_")])

    def __setattr__(self, name, value):
        self[name] = value

    def __getattr__(self, name):
        try:
            return self[name]
        except KeyError:
            raise AttributeError(f"'{self.__class__.__name__}' object has no attribute '{name}'")

    def dump(self, fp, *args, **kwargs):
        self._yaml.dump(self, fp)


def recursive_update(to_update, update_from):
    """
    Update values in `to_update` with values in `update_from`.

    Does not mutate values in to_update, but returns a new dictionary.
    """
    d = copy.deepcopy(to_update)
    for k, v in update_from.items():
        if isinstance(v, collections.abc.Mapping):
            d[k] = recursive_update(d.get(k, {}), v)
        else:
            d[k] = v
    return d


def which(file):
    # http://stackoverflow.com/questions/5226958/which-equivalent-function-in-python
    if os.path.exists(os.path.dirname(sys.executable) + "/" + file):
        return os.path.dirname(sys.executable) + "/" + file
    for path in os.environ["PATH"].split(":"):
        if os.path.exists(path + "/" + file):
            return path + "/" + file
    return None


def settings_to_sample():
    schema = Settings.schema_json()
    # expand schema for easier processing
    data = jsonref.loads(schema)
    strings = [process_property("gravity", data)]
    for key, value in data["properties"].items():
        strings.append(process_property(key, value, 1))
    concat = "\n".join(strings)
    return concat


def process_property(key, value, depth=0):
    extra_white_space = "  " * depth
    default = value.get("default", "")
    if isinstance(default, dict):
        # Little hack that prevents listing the default value for tusd in the sample config
        default = {}
    if default != "":
        # make values more yaml-like.
        default = yaml.dump(default)
        if default.endswith("\n...\n"):
            default = default[: -(len("\n...\n"))]
        default = default.strip()
    description = "\n".join(f"{extra_white_space}# {desc}" for desc in value["description"].strip().split("\n"))
    allOff = value.get("allOf", [])
    if allOff and allOff[0].get("properties"):
        # we've got a nested map, add key once
        description = f"{description}\n{extra_white_space}{key}:\n"
    has_child = False
    for item in allOff:
        if "enum" in item:
            description = f'{description}\n{extra_white_space}# Valid options are: {", ".join(item["enum"])}'
        if "properties" in item:
            has_child = True
            for _key, _value in item["properties"].items():
                description = f"{description}\n{process_property(_key, _value, depth=depth+1)}"
    if not has_child or key == "handlers":
        comment = "# "
        if key == "gravity":
            # gravity section should not be commented
            comment = ""
        if default == "":
            value_sep = ""
        else:
            value_sep = " "
        description = f"{description}\n{extra_white_space}{comment}{key}:{value_sep}{default}\n"
    return description



import click

from gravity import process_manager


NO_START_COMMANDS = (
    "shutdown",
    "status",
)


@click.command("supervisorctl")
@click.argument("supervisorctl_arg", nargs=-1)
@click.pass_context
def cli(ctx, supervisorctl_arg):
    """Invoke supervisorctl directly."""
    start_daemon = (supervisorctl_arg and supervisorctl_arg[0] not in NO_START_COMMANDS)
    with process_manager.process_manager(state_dir=ctx.parent.state_dir, start_daemon=start_daemon) as pm:
        pm.supervisorctl(*supervisorctl_arg)



import click

from gravity import options
from gravity import process_manager


@click.command("graceful")
@options.required_instance_arg()
@click.pass_context
def cli(ctx, instance):
    """Gracefully reload configured services.

    If INSTANCE matches an instance name, all services configured for the instance are restarted.

    If INSTANCE does not match an instance name, it is assumed to be a service and only the listed service(s) are
    restarted."""
    with process_manager.process_manager(state_dir=ctx.parent.state_dir) as pm:
        pm.graceful(instance)
